#BAR implementation
#Written by Stephen Lillington
import numpy as np
from scipy.optimize import root_scalar
import os
def read_xvg(xvgfile):
    with open(xvgfile,'r') as data:
        lines = data.readlines()

    i = 0
    #print(lines)
    while i < len(lines) and lines[i].startswith('@') or lines[i].startswith('#'):
        i += 1
    return np.loadtxt(xvgfile,skiprows=i)

def BAR_GROMACS(traj_lambda_state, nblocks, traj_file_path, Tol=1.0e-12, MaxIter=10000):
    state0_file = traj_file_path + "md" + str(traj_lambda_state) + ".xvg"
    state1_file = traj_file_path + "md" + str(traj_lambda_state+1) + ".xvg"
    if not os.path.isfile(state0_file) and os.path.isfile(state1_file):
        raise ValueError("File md$i.xvg does not exist in path %s" %(traj_lambda_state, os.getcwd()))
    state0_data = read_xvg(state0_file)
    state1_data = read_xvg(state1_file)
    blocked_state0_data = []
    block_indices = np.linspace(0,np.shape(state0_data)[0]-1,nblocks)

    def bar_fxn(dF):
        Nvals = -1.*deltaH_state0_l1_minus_l0 + dF
        Dvals = 1*deltaH_state1_l1_minus_l0 - dF
        maxterm = -max(Nvals.max(), Dvals.max())
        Num = np.sum(1. / (np.exp(Nvals + maxterm) + np.exp(maxterm)))
        Den = np.sum(1. / (np.exp(Dvals + maxterm) + np.exp(maxterm)))
        #rhs = np.log(Num / Den)
        return Num - Den
    #Should be the same for state1 data
    dFvals = np.zeros(nblocks-1)
    #dFvals = []
    #print("dF array" + str(type(dF)))
    index = 0
    for (i,j) in [(0,1),(1,2),(2,3),(3,4)]:
        start = int(block_indices[i])
        end = int(block_indices[j])
        state0_data_cut = state0_data[start:end,:]
        state1_data_cut = state1_data[start:end,:]

        if traj_lambda_state == 0:
            #Want data in column 6
            deltaH_state0_l1_minus_l0 = state0_data_cut[:,5]
            deltaH_state1_l1_minus_l0 = -state1_data_cut[:,4]

            #print("STATE DATA VECTORS")
            #print(deltaH_state0_l1_minus_l0,deltaH_state1_l1_minus_l0)
        elif traj_lambda_state == 1:
            deltaH_state0_l1_minus_l0 = state0_data_cut[:,6]
            deltaH_state1_l1_minus_l0 = -state1_data_cut[:,4]
            #print("STATE DATA VECTORS")
            #print(deltaH_state0_l1_minus_l0,deltaH_state1_l1_minus_l0)
        else:
            deltaH_state0_l1_minus_l0 = state0_data_cut[:,6]
            deltaH_state1_l1_minus_l0 = -state1_data_cut[:,4]
            #print("STATE DATA VECTORS")
            #print(deltaH_state0_l1_minus_l0,deltaH_state1_l1_minus_l0)

    #The current lambda state is column 7. (lambda - 1) is 6. (lambda + 1) is 8
    #But for last lambda state Col 8 is PV but current lambda is Col 7.
    #For lambda = 0, the current state is 5 and 0.1 is 6

        #Initial guess for dF
        #dU_traj0 = u_kln[0,1,nskip:] - u_kln[0,0,nskip:]
        #dU_traj1 = u_kln[1,1,nskip:] - u_kln[1,0,nskip:]
        F0 = -np.log(np.mean(np.exp(deltaH_state0_l1_minus_l0 - deltaH_state0_l1_minus_l0.max()))) - deltaH_state0_l1_minus_l0.max()
        F1 = np.log(np.mean(np.exp(deltaH_state1_l1_minus_l0 - deltaH_state1_l1_minus_l0.max()))) + deltaH_state1_l1_minus_l0.max()
        dF_0 = F1 - F0
        print("Initial guess: dF = %.8f" %dF_0)

        #Assuming that an equal number of samples were collected for every state

        #traj0_vals = u_kln[0,1,nskip:] - u_kln[0,0,nskip:]
        #traj1_vals = u_kln[1,1,nskip:] - u_kln[1,0,nskip:]

        #Set up initial points
        dF = dF_0
        Nvals = -1.*deltaH_state0_l1_minus_l0 + dF #-beta*dU + beta*dF
        Dvals = deltaH_state1_l1_minus_l0 - dF #beta*dU - beta*dF

        maxterm = -max(Nvals.max(), Dvals.max()) #For numeric stability
        Num = np.sum(1. / (np.exp(Nvals + maxterm) + np.exp(maxterm)))
        Den = np.sum(1. / (np.exp(Dvals + maxterm) + np.exp(maxterm)))

        #rhs is the right hand side of the eqn as written above in the BAR theory
        rhs = np.log(Num / Den)
        dFvals[index] = root_scalar(bar_fxn, method='bisect',bracket=[-20,20],x0=dF_0, xtol=Tol, maxiter=MaxIter).root
        index += 1

    #Compute variance in dF estimate
    avdF = np.mean(dFvals)
    stddev = np.std(dFvals)
    return avdF, stddev


def BAR_FreeEnergy(u_kln, nskip, Tol=1.0e-12, MaxIter=10000):
    """
    Estimates free energy difference between two states given a 3D array
    of potential energies (k x l x n). k = state for which dynamics progressed.
    l = potential energy computed for configuration from kth trajectory using
    the PE function l (e.g. lambda = lth value). n = sample for averaging.
    Inputs:
    ------
    u_kln = 3D Numpy array of reduced energies needed to compute a free energy difference.
    *reduced potential u_kln = beta * (u_l(x_kn) + P_l*V(x_kn) + mu_l'*N(x_kn))
    Has dimensions [2,2,Nsamples].
    Tol = Tolerance for convergence of free energy estimate
    MaxIter = Maximum number of iterations to converge on free energy estimated

    Outputs:
    --------
    Free energy difference
    Error in free energy difference estimate

    BAR theory
    Beta*dF = ln(mean(1./(1+exp(-Beta*(U01-U00)+Beta*dF))) / mean(1+exp(Beta*(U11-U10)-Beta*dF))))
    Numerator is an average of samples from the state 0 trajectory
    Denominator is an average of samples from the state 1 trajectory

    Iterate to solve for dF.

    **Potential energies are dimensionless (divided by kT) so we can leave out the Betas
    """
    '''
    def bar_fxn(dF):
        Nvals = -1.*traj0_vals + dF
        Dvals = traj1_vals - dF
        maxterm = -max(Nvals.max(), Dvals.max())
        Num = np.sum(1. / (np.exp(Nvals + maxterm) + np.exp(maxterm)))
        Den = np.sum(1. / (np.exp(Dvals + maxterm) + np.exp(maxterm)))
        rhs = np.log(Num / Den)
        return rhs - dF
    '''
    def bar_fxn(dF):
        Nvals = -1.*traj0_vals + dF
        Dvals = traj1_vals - dF
        maxterm = -max(Nvals.max(), Dvals.max())
        Num = np.sum(1. / (np.exp(Nvals + maxterm) + np.exp(maxterm)))
        Den = np.sum(1. / (np.exp(Dvals + maxterm) + np.exp(maxterm)))
        #rhs = np.log(Num / Den)
        return Num - Den

    #Initial guess for dF
    dU_traj0 = u_kln[0,1,nskip:] - u_kln[0,0,nskip:]
    dU_traj1 = u_kln[1,1,nskip:] - u_kln[1,0,nskip:]
    F0 = -np.log(np.mean(np.exp(dU_traj0 - dU_traj0.max()))) - dU_traj0.max()
    F1 = np.log(np.mean(np.exp(dU_traj1 - dU_traj1.max()))) + dU_traj1.max()
    dF_0 = F1 - F0
    print("Initial guess: dF = %.3f" %dF_0)

    #Assuming that an equal number of samples were collected for every state

    traj0_vals = u_kln[0,1,nskip:] - u_kln[0,0,nskip:]
    traj1_vals = u_kln[1,1,nskip:] - u_kln[1,0,nskip:]

    #Set up initial points
    dF = dF_0
    Nvals = -1.*traj0_vals + dF #-beta*dU + beta*dF
    Dvals = traj1_vals - dF #beta*dU - beta*dF

    maxterm = -max(Nvals.max(), Dvals.max()) #For numeric stability
    Num = np.sum(1. / (np.exp(Nvals + maxterm) + np.exp(maxterm)))
    Den = np.sum(1. / (np.exp(Dvals + maxterm) + np.exp(maxterm)))

    '''
    def bar_fxn(dF):
        Nvals = -1.*traj0_vals + dF
        Dvals = traj1_vals - dF
        maxterm = -max(Nvals.max(), Dvals.max())
        Num = np.sum(1. / (np.exp(Nvals + maxterm) + np.exp(maxterm)))
        Den = np.sum(1. / (np.exp(Dvals + maxterm) + np.exp(maxterm)))
        rhs = np.log(Num / Den)
        return rhs - dF
    '''
    #rhs is the right hand side of the eqn as written above in the BAR theory
    rhs = np.log(Num / Den)
    dF = root_scalar(bar_fxn, method='bisect',bracket=[-5,5],x0=dF_0, xtol=Tol, maxiter=MaxIter)
    return dF.root
